#include <QtGui>
#include "mainform.h"
#include "channel.h"
#include "gdata.h"
#include "soundfile.h"
#include <QMessageBox>
#include "main.h"
#include "tartinisettingsdialog.h"
#include "ui_annotation.h"
#include "ui_about.h"
#include <QTextBrowser>

MainForm::MainForm(QMainWindow *parent) : QMainWindow(parent)
{
  //
  //  Open the UI window.
  //
  ui.setupUi(this);
  tmp_filename = NULL;
  tmp_soundfile = NULL;
  //
  //  Set the window icon
  //
  QIcon *i = new QIcon(":icons/icon.bmp");
  QWidget::setWindowIcon(*i);
}

MainForm::~MainForm()
{
  //
  //  Delete the temp record file if it is still there.
  //
  if (tmp_soundfile) {
    tmp_filename = new char[strlen(tmp_soundfile->filename)];
    (void) strcpy(tmp_filename, tmp_soundfile->filename);
    tmp_soundfile->uninit();
    ::remove(tmp_filename);
    delete [] tmp_filename;
    tmp_filename = NULL;
    delete tmp_soundfile;
    tmp_soundfile = NULL;
  };
  //
  //  Save Geometry
  //
  gdata->qsettings->setValue("mainForm/geometry", saveGeometry());
};

void MainForm::init(QObject *app)
{
  connect(ui.resumeButton, SIGNAL(clicked()), this, SLOT(openRecord()));
  connect(qApp, SIGNAL(QApplication::lastWindowClose()), this, SLOT(Exit()));

  connect(ui.actionExit, SIGNAL(triggered()), this, SLOT(Exit()));
  connect(ui.actionAnalyze, SIGNAL(triggered()), this, SLOT(Analyze()));
  connect(ui.actionExport, SIGNAL(triggered()), this, SLOT(Export()));
  connect(ui.actionPrint, SIGNAL(triggered()), this, SLOT(Print()));
  connect(ui.actionPreferences, SIGNAL(triggered()), this, SLOT(Preferences()));
  connect(ui.actionSummary, SIGNAL(triggered()), this, SLOT(Summary()));
  connect(ui.actionAbout, SIGNAL(triggered()), this, SLOT(About()));
  connect(ui.actionHelp, SIGNAL(triggered()), this, SLOT(Help()));

  //
  //  Restore Geometry
  //
  if (gdata->qsettings->contains("mainForm/geometry")) {
      restoreGeometry(gdata->qsettings->value("mainForm/geometry").toByteArray());
    };
};

void displayInBrowser(const QString& theURL)
{
  ::ShellExecute(qApp->mainWidget()->winId(),NULL,theURL,NULL,NULL,SW_SHOW);
};

void MainForm::Help()
{
  displayInBrowser(QString("http://www.novasession.org/Flutini/tutorial.html"));
};

void MainForm::About()
{
  QDialog *aboutForm = new QDialog;
  Ui_aboutDialog af_ui;
  af_ui.setupUi(aboutForm);
  aboutForm->exec();
};

void MainForm::Exit()
{
#ifdef DEBUG_MEMORY
  DumpUnfreed(TRUE);
#endif
  if(gdata->running) {
    gdata->stop();
  }
  close();
};

void MainForm::Print()
{
  QPrinter printer;
  QString annText;

  if (gdata->getPrintAnnotate()) {
    QDialog *annForm = new QDialog;
    Ui_AnnotationForm af_ui;
    af_ui.setupUi(annForm);
    annForm->exec();
    annText = af_ui.annotationText->toPlainText();
  };

  QPrintDialog *dialog = new QPrintDialog(&printer, this);
  dialog->setWindowTitle(tr("Print Display"));
  if (dialog->exec() != QDialog::Accepted) {
    delete dialog;
    return;
  };
  delete dialog;

  QPixmap qp1 = QPixmap::grabWidget(ui.RTTA_display);
  QPixmap qp2 = QPixmap::grabWidget(ui.bar_display);

  QPainter painter;
  QFont title_font("Courier", 10, QFont::Bold);
  QFont body_font("Courier", 10);
  painter.begin(&printer);
  painter.drawPixmap(5,5,qp1);
  painter.drawPixmap(210,5,qp2);

  //
  //  Annotation text if necessary
  //
  if (gdata->getPrintAnnotate() && !annText.isEmpty()) {
    painter.setBackgroundMode(Qt::TransparentMode);
    painter.setPen("black");
    painter.setFont(title_font);
    painter.drawText(5, qp1.height()+20, "NOTES:");
    painter.drawRect(5, qp1.height()+25, qp1.width() + qp2.width(), 400);
    painter.setFont(body_font);
    painter.drawText(10, qp1.height()+35, qp1.width() + qp2.width() - 5, 390, Qt::AlignLeft & Qt::AlignTop & Qt::TextWordWrap, annText);
  };
  
  painter.end();

};

void MainForm::Analyze()
{
  Reset();
  //
  //  Turning on disk files?
  //
  gdata->no_disk_files = true;
  
  openFile();
};

void MainForm::Summary()
{
  QString annText;

  if (gdata->getSummaryAnnotate()) {
    QDialog *annForm = new QDialog;
    Ui_AnnotationForm af_ui;
    af_ui.setupUi(annForm);
    annForm->exec();
    annText = af_ui.annotationText->toPlainText();
  };

  QString s = QFileDialog::getSaveFileName(mainWindow, "Choose a filename to save under", ".", "*.txt");
  if(s.isEmpty()) return;

  QFile f(s);
  f.open(QIODevice::WriteOnly);
  QTextStream out(&f);

  out << "Flutini 1.0" << endl;
  out << "Total samples: " << gdata->rtta_data->total_samples << endl; 
  out << qSetFieldWidth(18);
  out << "Note" << "Samples" << "Delta in Cents" << endl;
  //
  //  Print the bins
  //
  for(int i=0; i<RTTA_NOTE_RANGE; i++) {
    if (gdata->rtta_data->rttaInfo[i].num_samples)
      out << gdata->rtta_data->rttaInfo[i].note_name
	  << gdata->rtta_data->rttaInfo[i].num_samples
	  << (int) (((gdata->rtta_data->rttaInfo[i].total/gdata->rtta_data->rttaInfo[i].num_samples) - gdata->rtta_data->rttaInfo[i].start_val - 0.50) * 100) << endl;
  }

    //
  //  Annotation text if necessary
  //
  if (gdata->getSummaryAnnotate() && !annText.isEmpty()) {
    out << "NOTES:" << endl << endl;
    out << annText;
  };

  f.close();
};


void MainForm::Export()
{
  QString s = QFileDialog::getSaveFileName(mainWindow, "Choose a filename to save under", ".", "*.txt");
  if(s.isEmpty()) return;

  QFile f(s);
  f.open(QIODevice::WriteOnly);
  QTextStream out(&f);

  //
  // Naive attempt to get the old text dump back.  Doesn't work... Dies
  // trying to get the size of saved chunks in the channel, probably
  // because I'm not saving anymore?
  //
  if (gdata->getActiveChannel()) {
    gdata->getActiveChannel()->exportChannel(out);
  } else {
    out << "No detailed information available." << endl;
  };
  
  f.close();
};

void MainForm::Preferences()
{
  TartiniSettingsDialog *settings = new TartiniSettingsDialog(this);
  settings->setUnknownsToDefault(gdata->qsettings);
  gdata->qsettings->sync();
  settings->init();
  settings->show();
};

void MainForm::Reset()
{
  gdata->rtta_data->reset();
};
void MainForm::Resume()
{
};

void MainForm::openRecord()
{
  //
  //  Turning off disk files
  //
  gdata->no_disk_files = true;
  
  //
  //  This toggles ... is that what we want?
  //
  if(gdata->running) {
    gdata->stop();
    ui.resumeButton->setText("Start");
#ifdef SRT
    //
    //  Delete the old file...
    //
    if (tmp_soundfile) {
      tmp_filename = new char[strlen(tmp_soundfile->filename)];
      (void) strcpy(tmp_filename, tmp_soundfile->filename);
      tmp_soundfile->uninit();
      ::remove(tmp_filename);
      delete [] tmp_filename;
      tmp_filename = NULL;
      delete tmp_soundfile;
      tmp_soundfile = NULL;
    };
#endif
    return;
  }

  //
  //  Delete the old file if there is one.  
  //
  if (tmp_soundfile) {
    tmp_filename = new char[strlen(tmp_soundfile->filename)];
    (void) strcpy(tmp_filename, tmp_soundfile->filename);
    tmp_soundfile->uninit();
    ::remove(tmp_filename);
    delete [] tmp_filename;
    tmp_filename = NULL;
    delete tmp_soundfile;
    tmp_soundfile = NULL;
  };

  //
  //  Reset the RTTA stuff.
  //
  Reset();

  ui.resumeButton->setText("Stop");

  //
  //  We don't play at the same time...
  //
  SoundFile *playSoundFile = NULL;
  int rate = gdata->qsettings->value("Sound/sampleRate", 44100).toInt();
  QString numberOfChannels =
    gdata->qsettings->value("Sound/numberOfChannels", "mono").toString();
  int channels;
  if (numberOfChannels.lower() == "mono") channels = 1;
  else channels = 2;

  int bits = gdata->qsettings->value("Sound/bitsPerSample", 16).toInt();
  int windowSize = gdata->getAnalysisBufferSize(rate);
  int stepSize = gdata->getAnalysisStepSize(rate);

  QString tempFileFolder = gdata->qsettings->value("General/tempFilesFolder",
						   QDir::convertSeparators(QDir::currentDirPath())).toString();

  QDir dir = QDir(tempFileFolder);
  QFileInfo fileInfo; //(dir, fileName);
  QString fileName;
  bool fileExists;
  do {
    fileExists = false;
    fileName = gdata->getFilenameString();
    fileInfo.setFile(dir, fileName);
    if(fileInfo.exists()) {
      fileExists = true;
      int fileGeneratingNumber = gdata->qsettings->value("General/fileGeneratingNumber", 1).toInt();
      gdata->qsettings->setValue("General/fileGeneratingNumber", fileGeneratingNumber+1);
    }
  } while(fileExists);
  
  SoundFile *newSoundFile = new SoundFile();
  QString newFilename(fileInfo.absFilePath().latin1());
  newFilename = QDir::convertSeparators(newFilename);
  //
  //  SRT -- Here's where we open the recording file.
  //
  if(!newSoundFile->openWrite(newFilename, rate, channels, bits, windowSize, stepSize)) {
    delete newSoundFile; newSoundFile = NULL;
    return;
  }
  tmp_soundfile = newSoundFile;
  
  if(!gdata->openPlayRecord(newSoundFile, playSoundFile)) {
    // gdata->closeFile(newSoundFile, GData::NEVER_SAVE/*, false*/);
    QMessageBox::warning(mainWindow, "Error", QString("Error opening sound device for recording"), QMessageBox::Ok, QMessageBox::NoButton);
    return;
  }
  gdata->addFileToList(newSoundFile);
  gdata->view->setCurrentTime(0.0);
  gdata->setActiveChannel(newSoundFile->channels(0));

  int fileGeneratingNumber = gdata->qsettings->value("General/fileGeneratingNumber", 1).toInt();
  gdata->qsettings->setValue("General/fileGeneratingNumber", fileGeneratingNumber+1);
}

void MainForm::openFile()
{
  QString lastFolder = QDir::convertSeparators(gdata->qsettings->value("Dialogs/openFilesFolder", QDir::currentDirPath()).toString());
  QString s = QFileDialog::getOpenFileName(this, "Open File", lastFolder, "Sounds (*.wav)");
  if(s.isEmpty()) return;
  s = QDir::convertSeparators(s);
  gdata->qsettings->setValue("Dialogs/openFilesFolder", s);
  const char *filename = s.latin1();
  SoundFile *newSoundFile = new SoundFile();

  //
  //  Delete the old file if there is one.  
  //
  if (tmp_soundfile) {
    tmp_filename = new char[strlen(tmp_soundfile->filename)];
    (void) strcpy(tmp_filename, tmp_soundfile->filename);
    tmp_soundfile->uninit();
    ::remove(tmp_filename);
    delete [] tmp_filename;
    tmp_filename = NULL;
    delete tmp_soundfile;
    tmp_soundfile = NULL;
  };

  if(!newSoundFile->openRead(filename)) {
    fprintf(stderr, "Error opening %s\n", filename);
    delete newSoundFile;
    return;
  }
  
  newSoundFile->preProcess();
  gdata->updateViewLeftRightTimes();
  
  gdata->addFileToList(newSoundFile);
  gdata->setActiveChannel(newSoundFile->channels(0));
  QApplication::postEvent(mainWindow, new QCustomEvent(UPDATE_SLOW));
  gdata->view->doUpdate();

  // Clean up this soundfile
  // delete newSoundFile;
}

